package org.example

data class State(
    val input: String,
    val startIndex: Int,
    val endIndex: Int,
    val captures: Map<String, String> = emptyMap()
) {
  val matched: String
    get() = input.substring(startIndex, endIndex)
  val remaining: String
    get() = input.substring(endIndex)
}

typealias AvailableState = Sequence<State>

// 재귀 하향 분석기.
// 백트랙킹 기반.
interface RegexItem {
  override fun toString(): String
  fun findMatch(str: String, position: Int = 0): AvailableState
}

fun RegexItem.findAll(item: String): Sequence<State> {
  return sequence {
    var position = 0
    // 문자열의 끝까지 반복합니다. 비어있어도 한번은 시도합니다.
    while (position <= item.length) {
      // findMatch 메서드를 호출하여 매칭을 시도합니다.
      val matchResult = findMatch(item, position).firstOrNull()
      if (matchResult == null) {
        // 매칭이 실패하면 position을 증가시키고 다시 시도합니다.
        position++
        continue
      }
      // 매칭이 성공하면 MatchResult를 생성하여 반환합니다.
      yield(matchResult)

      // 다음 위치로 이동합니다.
      position =
          if (matchResult.startIndex == matchResult.endIndex) {
            position + 1
          } else {
            matchResult.endIndex
          }
    }
  }
}

fun RegexItem.find(item: String): State? {
  // findAll 에서 첫 번째 매칭 결과를 반환합니다.
  return this.findAll(item).firstOrNull()
}

fun RegexItem.containsMatchIn(item: String): Boolean {
  // 매칭 결과가 성공인지 확인하는 헬퍼 함수
  return this.find(item) != null
}

class AndThenItem(val left: RegexItem, val right: RegexItem) : RegexItem {
  override fun toString(): String = "${left}${right}"
  override fun findMatch(str: String, position: Int): AvailableState {
    val leftMatch = left.findMatch(str, position)
    return leftMatch.flatMap { leftState ->
      right.findMatch(str, leftState.endIndex).map { rightState ->
        // If right match is successful, combine the matched parts
        State(
            str,
            leftState.startIndex,
            rightState.endIndex,
            leftState.captures + rightState.captures // Combine captures
        )
      }
    }
  }
}

class CharItem(val value: String) : RegexItem {
  override fun toString(): String =
      // escape 특수 문자를 처리하여 출력
      when (value) {
        "\\", "+", "*", "?", ".", "(", ")", "|", "[", "]" -> "\\$value"
        else -> value // 일반 문자 그대로 반환
      }

  override fun findMatch(str: String, position: Int): AvailableState {
    return when {
      // 첫번째 문자가 value와 일치하는지 확인
      position < str.length && str[position].toString() == value -> {
        sequenceOf(State(str, position, position + 1))
      }
      else -> emptySequence() // 일치하지 않으면 빈 시퀀스 반환
    }
  }
}

class BracketItem(val content: String) : RegexItem {
  override fun toString(): String = "[$content]"

  // TODO: 범위 처리
  override fun findMatch(str: String, position: Int): AvailableState {
    // 대괄호 안의 내용과 일치하는 첫 문자를 찾음
    return when {
      position < str.length && content.contains(str[position]) -> {
        sequenceOf(State(str, position, position + 1))
      }
      else -> emptySequence() // 일치하지 않으면 빈 시퀀스 반환
    }
  }
}

class GroupItem(val item: RegexItem, val name: String) : RegexItem {
  override fun toString(): String = "(${item})"

  override fun findMatch(str: String, position: Int): AvailableState {
    // 그룹은 내부 아이템과 동일하게 매칭을 시도
    val ret = item.findMatch(str, position)
    // 매칭된 상태에 그룹 이름을 추가하여 반환
    return ret.map { state -> state.copy(captures = state.captures + (name to state.matched)) }
  }
}

class AnchorItem(val anchor: String) : RegexItem {
  override fun toString(): String = anchor

  override fun findMatch(str: String, position: Int): AvailableState {
    // 앵커는 문자열의 시작(^) 또는 끝($)과 매칭됨
    return when (anchor) {
      "^" ->
          if (position == 0) {
            sequenceOf(State(str, 0, 0))
          } else {
            emptySequence() // 시작 앵커가 실패하면 빈 시퀀스 반환
          }
      "$" ->
          if (position == str.length) {
            sequenceOf(State(str, str.length, str.length))
          } else {
            emptySequence() // 끝 앵커가 실패하면 빈 시퀀스 반환
          }
      // 다른 앵커는 지원하지 않음
      else -> throw IllegalArgumentException("Unknown anchor: $anchor")
    }
  }
}

fun matchMany(str: String, item: RegexItem, position: Int): Sequence<State> {
  // 욕심쟁이 매칭을 위한 헬퍼 함수
  return item.findMatch(str, position).flatMap { state ->
    if (state.endIndex == str.length) {
      sequenceOf(state) // If remaining is empty, return the matched state
    } else {
      // Otherwise, continue matching with the remaining string
      matchMany(str, item, state.endIndex).map { nextState ->
        State(str, state.startIndex, nextState.endIndex)
      } + sequenceOf(state) // Include the current state as well
    }
  }
}

class PlusItem(val item: RegexItem) : RegexItem {
  override fun toString(): String = "${item}+"
  override fun findMatch(str: String, position: Int): AvailableState {
    return matchMany(str, item, position)
  }
}

class StarItem(val item: RegexItem) : RegexItem {
  override fun toString(): String = "${item}*"
  override fun findMatch(str: String, position: Int): AvailableState {
    // *는 0번 또는 1번 이상 일치합니다.
    // 욕심쟁이(greedy) 방식으로 구현하기 위해, 가장 긴 매치(1번 이상)를 먼저 찾고, 그 다음에 0번 매치를 추가합니다.
    val oneOrMoreMatches = matchMany(str, item, position)
    val zeroMatch = sequenceOf(State(str, position, position))
    return oneOrMoreMatches + zeroMatch
  }
}

class QuestionItem(val item: RegexItem) : RegexItem {
  override fun toString(): String = "${item}?"
  override fun findMatch(str: String, position: Int): AvailableState {
    // ?는 0번 또는 1번 일치합니다.
    val oneMatch = item.findMatch(str, position)
    val zeroMatch = sequenceOf(State(str, position, position))
    // 1번 매치를 0번 매치보다 우선합니다.
    return oneMatch + zeroMatch
  }
}

class DotItem : RegexItem {
  override fun toString(): String = "."
  override fun findMatch(str: String, position: Int): AvailableState =
      // .은 임의의 한 문자와 매칭되므로, 첫 문자가 존재하면 매칭 성공
      when {
        position < str.length -> sequenceOf(State(str, position, position + 1))
        else -> emptySequence() // 빈 문자열에 대해서는 매칭 실패
      }
}

class AlternationItem(val left: RegexItem, val right: RegexItem) : RegexItem {
  override fun toString(): String = "${left}|${right}"
  override fun findMatch(str: String, position: Int): AvailableState {
    // Alternation은 왼쪽 또는 오른쪽 항목 중 하나와 매칭되므로, 각각 시도해보고 성공하는 경우를 반환
    val leftMatch = left.findMatch(str, position)
    val rightMatch = right.findMatch(str, position)

    return leftMatch + rightMatch // 두 매칭 결과를 합쳐서 반환
  }
}